@@ -52,10 +52,12 @@ def nth(iterable, n, default=None):
     """
     return next(islice(iterable, n, None), default)
 
-# widely-spreaded shoutcuts to get first item 
-# and all but first item from iterator respectively
-head = partial(flip(nth), 0)
-tail = partial(drop, 1)
+# widely-spreaded shortcuts to get first item, all but first item,
+# second item, and first item of first item from iterator respectively
+head = first = partial(flip(nth), 0)
+tail = rest = partial(drop, 1)
+second = F(rest) >> first
+ffirst = F(first) >> first
 
 # shortcut to remove all falsey items from iterable
 compact = partial(filter, None)
@@ -66,6 +68,13 @@ def reject(func, iterable):
     """
     return filter(F(not_) << (func or _), iterable)
 
+def iterate(f, x):
+    """Return an iterator yielding x, f(x), f(f(x)) etc.
+    """
+    while True:
+        yield x
+        x = f(x)
+
 def padnone(iterable):
     """Returns the sequence elements and then returns None indefinitely.
     Useful for emulating the behavior of the built-in map() function.
@@ -100,6 +109,16 @@ def grouper(n, iterable, fillvalue=None):
     args = [iter(iterable)] * n
     return zip_longest(*args, fillvalue=fillvalue)
 
+def group_by(keyfunc, iterable):
+    """Returns a dict of the elements from given iterable keyed by result
+    of keyfunc on each element. The value at each key will be a list of
+    the corresponding elements, in the order they appeared in the iterable.
+    """
+    grouped = {}
+    for item in iterable:
+        grouped.setdefault(keyfunc(item), []).append(item)
+    return grouped
+
 def roundrobin(*iterables):
     """roundrobin('ABC', 'D', 'EF') --> A D E B F C
     Recipe originally credited to George Sakkis.
@@ -158,7 +177,7 @@ def pairwise(iterable):
     next(b, None)
     return zip(a, b)
 
-def iter_except(func, exception, first=None):
+def iter_except(func, exception, first_=None):
     """ Call a function repeatedly until an exception is raised.
 
     Converts a call-until-exception interface to an iterator interface.
@@ -175,8 +194,8 @@ def iter_except(func, exception, first=None):
     http://docs.python.org/3.4/library/itertools.html#itertools-recipes
     """
     try:
-        if first is not None:
-            yield first()            # For database APIs needing an initial cast to db.first()
+        if first_ is not None:
+            yield first_()            # For database APIs needing an initial cast to db.first()
         while 1:
             yield func()
     except exception:
