@@ -8,6 +8,8 @@ import operator
 import itertools
 
 from fn import op, _, F, Stream, iters, underscore, monad, recur
+from fn.uniform import reduce
+from fn.immutable import SkewHeap, PairingHeap, LinkedList, Stack, Queue, Vector, Deque
 
 class InstanceChecker(object):
     if sys.version_info[0] == 2 and sys.version_info[1] <= 6:
@@ -285,6 +287,8 @@ class UnderscoreTestCase(unittest.TestCase):
         self.assertEqual("(x1) => (x1 == 2)", str(_ == 2))
         self.assertEqual("(x1) => (x1 != 2)", str(_ != 2))
 
+        self.assertEqual("(x1) => ((x1 * 2) + 1)", str((_ * 2 + 1)))
+
     def test_rigthside_string_converting(self):
         self.assertEqual("(x1) => (2 + x1)",  str(2 + _))
         self.assertEqual("(x1) => (2 - x1)",  str(2 - _))
@@ -327,10 +331,14 @@ class UnderscoreTestCase(unittest.TestCase):
         self.assertEqual("(x1, x2) => (x1 == x2)", str(_ == _))
         self.assertEqual("(x1, x2) => (x1 != x2)", str(_ != _))
 
+        self.assertEqual("(x1, x2) => (((x1 / x2) - 1) * 100)", str((_ / _ - 1) * 100))
+
     def test_reverse_string_converting(self):
         self.assertEqual("(x1, x2, x3) => ((x1 + x2) + x3)", str(_ + _ + _))
         self.assertEqual("(x1, x2, x3) => (x1 + (x2 * x3))", str(_ + _ * _))
 
+        self.assertEqual("(x1) => (1 + (2 * x1))", str((1 + 2 * _)))
+
     def test_multi_underscore_string_converting(self):
         self.assertEqual("(x1) => (x1 + '_')", str(_ + "_"))
         self.assertEqual("(x1, x2) => getattr((x1 + x2), '__and_now__')", str((_ + _).__and_now__))
@@ -341,6 +349,13 @@ class UnderscoreTestCase(unittest.TestCase):
         self.assertEqual(_ + _, eval(repr(_ + _)))
         self.assertEqual(_ + _ * _, eval(repr(_ + _ * _)))
 
+    def test_repr_parse_str(self):
+        self.assertEqual('=> ' + _, eval(repr('=> ' + _)))
+        self.assertEqual(
+            reduce(lambda f, n: f.format(n), ('({0} & _)',) * 11).format('_'),
+            repr(reduce(_ & _, (_,) * 12)),
+        )
+
 class CompositionTestCase(unittest.TestCase):
 
     def test_composition(self):
@@ -419,6 +434,9 @@ class IteratorsTestCase(unittest.TestCase):
 
         self.assertEqual(1, iters.head(gen()))
 
+    def test_first(self):
+        self.assertEqual(iters.first, iters.head)  # Check if same object
+
     def test_tail(self):
         self.assertEqual([1,2], list(iters.tail([0,1,2])))
         self.assertEqual([], list(iters.tail([])))
@@ -430,6 +448,29 @@ class IteratorsTestCase(unittest.TestCase):
 
         self.assertEqual([2,3], list(iters.tail(gen())))
 
+    def test_rest(self):
+        self.assertEqual(iters.rest, iters.tail)  # Check if same object
+
+    def test_second(self):
+        self.assertEqual(2, iters.second([1, 2, 3]))
+        self.assertEqual(None, iters.second([]))
+
+        def gen():
+            yield 10
+            yield 20
+            yield 30
+
+        self.assertEqual(20, iters.second(gen()))
+
+    def test_ffirst(self):
+        self.assertEqual(1, iters.ffirst([[1, 2], [3, 4]]))
+        self.assertEqual(None, iters.ffirst([[], [10, 20]]))
+
+        def gen():
+            yield (x * 10 for x in (10, 20, 30,))
+
+        self.assertEqual(100, iters.ffirst(gen()))
+
     def test_compact(self):
         self.assertEqual([True, 1, 0.1, "non-empty", [""], (0,), {"a": 1}],
                          list(iters.compact([None, False, True, 0, 1, 0.0, 0.1,
@@ -445,6 +486,13 @@ class IteratorsTestCase(unittest.TestCase):
                                                   [], [""], (), (0,),
                                                   {}, {"a": 1}])))
 
+    def test_iterate(self):
+        it = iters.iterate(lambda x: x * x, 2)
+        self.assertEqual(2, next(it))  # 2
+        self.assertEqual(4, next(it))  # 2 * 2
+        self.assertEqual(16, next(it))  # 4 * 4
+        self.assertEqual(256, next(it))  # 16 * 16
+
     def test_padnone(self):
         it = iters.padnone([10,11])
         self.assertEqual(10, next(it))
@@ -489,6 +537,19 @@ class IteratorsTestCase(unittest.TestCase):
         self.assertEqual(["D","E","F"], list(b))
         self.assertEqual(["G","x","x"], list(c))
 
+    def test_group_by(self):
+        # verify grouping logic
+        grouped = iters.group_by(len, ['1', '12', 'a', '123', 'ab'])
+        self.assertEqual({1: ['1', 'a'], 2: ['12', 'ab'], 3: ['123']}, grouped)
+
+        # verify it works with any iterable - not only lists
+        def gen():
+            yield '1'
+            yield '12'
+
+        grouped = iters.group_by(len, gen())
+        self.assertEqual({1: ['1'], 2: ['12']}, grouped)
+
     def test_roundrobin(self):
         r = iters.roundrobin('ABC', 'D', 'EF')
         self.assertEqual(["A","D","E","B","F","C"], list(r))
@@ -775,5 +836,267 @@ class TrampolineTestCase(unittest.TestCase):
 
         self.assertEqual(5000, recur_inc2(10000))
 
+class UnionBasedHeapsTestCase(unittest.TestCase):
+
+    def _heap_basic_operations(self, cls):
+        # Create new heap with 3 elements
+        s1 = cls(10)
+        s2 = s1.insert(30)
+        s3 = s2.insert(20)
+
+        # Extract elements one-by-one
+        el1, sx1 = s3.extract()
+        el2, sx2 = sx1.extract()
+        el3, sx3 = sx2.extract()
+
+        # Check elements ordering
+        self.assertEqual(10, el1)
+        self.assertEqual(20, el2)
+        self.assertEqual(30, el3)
+
+        # Check that previous heap are persistent
+        el22, _ = sx1.extract()
+        self.assertEqual(20, el22)
+
+    def _heap_iterator(self, cls):
+        # Create new heap with 5 elements
+        h = cls(10)
+        h = h.insert(30)
+        h = h.insert(20)
+        h = h.insert(5)
+        h = h.insert(100)
+
+        # Convert to list using iterator
+        self.assertEqual([5, 10, 20, 30, 100], list(h))
+
+    def _heap_custom_compare(self, cls):
+        h = cls(cmp=lambda a,b: len(a) - len(b))
+        h = h.insert("give")
+        h = h.insert("few words")
+        h = h.insert("about")
+        h = h.insert("union heaps")
+        h = h.insert("implementation")
+
+        # Convert to list using iterator
+        self.assertEqual(["give",
+                          "about",
+                          "few words",
+                          "union heaps",
+                          "implementation"], list(h))
+
+    def _heap_compare_with_keyfunc(self, cls):
+        from operator import itemgetter
+
+        # Create new heap with 5 elements
+        h = cls(key=itemgetter(1))
+        h = h.insert((10, 10))
+        h = h.insert((30, 15))
+        h = h.insert((20, 110))
+        h = h.insert((40, -10))
+        h = h.insert((50, 100))
+
+        # Convert to list using iterator
+        self.assertEqual([(40,-10), (10,10), (30,15), (50,100), (20,110)], list(h))
+
+    def test_skew_heap_basic(self):
+        self._heap_basic_operations(SkewHeap)
+
+    def test_pairing_heap_basic(self):
+        self._heap_basic_operations(PairingHeap)
+
+    def test_skew_heap_iterator(self):
+        self._heap_iterator(SkewHeap)
+
+    def test_pairing_heap_iterator(self):
+        self._heap_iterator(PairingHeap)
+
+    def test_skew_heap_key_func(self):
+        self._heap_compare_with_keyfunc(SkewHeap)
+
+    def test_pairing_heap_key_func(self):
+        self._heap_compare_with_keyfunc(PairingHeap)
+
+    def test_skew_heap_cmp_func(self):
+        self._heap_custom_compare(SkewHeap)
+
+    def test_pairing_heap_cmp_func(self):
+        self._heap_custom_compare(PairingHeap)
+
+class LinkedListsTestCase(unittest.TestCase):
+
+    def test_linked_list_basic_operations(self):
+        l1 = LinkedList()
+        l2 = l1.cons(1)
+        l3 = l2.cons(2)
+        self.assertEqual(None, l1.head)
+        self.assertEqual(1, l2.head)
+        self.assertEqual(2, l3.head)
+        self.assertEqual(1, l3.tail.head)
+        self.assertEqual(None, l3.tail.tail.head)
+
+    def test_linked_list_num_of_elements(self):
+        self.assertEqual(0, len(LinkedList()))
+        self.assertEqual(3, len(LinkedList().cons(10).cons(20).cons(30)))
+
+    def tests_linked_list_iterator(self):
+        self.assertEqual([30, 20, 10], list(LinkedList().cons(10).cons(20).cons(30)))
+    
+    def test_from_iterable(self):
+        expected = [10, 20, 30]
+        actual = list(LinkedList.from_iterable(expected))
+        self.assertEqual(actual, expected)
+
+        actual = LinkedList.from_iterable(tuple(expected))
+        self.assertEqual(list(actual), expected)
+
+        actual = LinkedList.from_iterable(iter(expected))
+        self.assertEqual(list(actual), expected)
+
+        actual = LinkedList.from_iterable(LinkedList().cons(30).cons(20).cons(10))
+        self.assertEqual(list(actual), expected)
+
+    def test_stack_push_pop_ordering(self):
+        s1 = Stack()
+        s2 = s1.push(1)
+        s3 = s2.push(10)
+        s4 = s3.push(100)
+        (sv4, s5) = s4.pop()
+        (sv3, s6) = s5.pop()
+        self.assertEqual(100, sv4)
+        self.assertEqual(10, sv3)
+        self.assertEqual(100, s4.pop()[0])
+
+    def test_stack_length(self):
+        self.assertEqual(0, len(Stack()))
+        self.assertEqual(3, len(Stack().push(1).push(2).push(3)))
+
+    def test_stack_is_empty_check(self):
+        self.assertTrue(Stack().push(100))
+        self.assertFalse(Stack().push(100).is_empty())
+        self.assertTrue(Stack().is_empty())
+
+    def test_pop_empty_stack_exception(self):
+        self.assertRaises(ValueError, Stack().pop)
+
+    def test_stack_iterator(self):
+        self.assertEqual([10, 5, 1], list(Stack().push(1).push(5).push(10)))
+        self.assertEqual(6, sum(Stack().push(1).push(2).push(3)))
+
+class BankerQueueTestCase(unittest.TestCase):
+    
+    def test_queue_basic_operations(self):
+        q1 = Queue()
+        q2 = q1.enqueue(1)
+        q3 = q2.enqueue(10)
+        q4 = q3.enqueue(100)
+        self.assertEqual(1, q4.dequeue()[0])
+        self.assertEqual(1, q3.dequeue()[0])
+        self.assertEqual(1, q2.dequeue()[0])
+        v1, q5 = q4.dequeue()
+        v2, q6 = q5.dequeue()
+        v3, q7 = q6.dequeue()
+        self.assertEqual(1, v1)
+        self.assertEqual(10, v2)
+        self.assertEqual(100, v3)
+        self.assertEqual(0, len(q7))
+
+    def test_queue_num_of_elements(self):
+        self.assertEqual(0, len(Queue()))
+        self.assertEqual(3, len(Queue().enqueue(1).enqueue(2).enqueue(3)))
+
+    def test_queue_is_empty(self):
+        self.assertTrue(Queue().is_empty())
+        self.assertFalse(Queue().enqueue(1).is_empty())
+        self.assertTrue(Queue().enqueue(1).dequeue()[1].is_empty())
+
+    def test_dequeue_from_empty(self):
+        self.assertRaises(ValueError, Queue().dequeue)
+
+    def test_iterator(self):
+        self.assertEqual([], list(Queue()))
+        self.assertEqual([1,2,3], list(Queue().enqueue(1).enqueue(2).enqueue(3)))
+        self.assertEqual(60, sum(Queue().enqueue(10).enqueue(20).enqueue(30)))
+
+class VectorTestCase(unittest.TestCase):
+
+    def test_cons_operation(self):
+        v = Vector()
+        self.assertEqual(0, len(v))
+        v1 = v.cons(10)
+        self.assertEqual(1, len(v1))
+        self.assertEqual(0, len(v)) # previous value didn't change
+        up = reduce(lambda acc, el: acc.cons(el), range(513), Vector())
+        self.assertEqual(513, len(up))
+
+    def test_assoc_get_operations(self):
+        v = Vector()
+        v1 = v.assoc(0, 10)
+        v2 = v1.assoc(1, 20)
+        v3 = v2.assoc(2, 30)
+        self.assertEqual(10, v3.get(0))
+        self.assertEqual(20, v3.get(1))
+        self.assertEqual(30, v3.get(2))
+        # check persistence
+        v4 = v2.assoc(2, 50)
+        self.assertEqual(30, v3.get(2))
+        self.assertEqual(50, v4.get(2))
+        # long vector
+        up = reduce(lambda acc, el: acc.assoc(el, el*2), range(1500), Vector())
+        self.assertEqual(2800, up.get(1400))
+        self.assertEqual(2998, up.get(1499))
+
+    def test_pop_operations(self):
+        v = reduce(lambda acc, el: acc.cons(el), range(2000), Vector())
+        self.assertEqual(1999, len(v.pop()))
+        self.assertEqual(list(range(1999)), list(v.pop()))
+
+    def test_vector_iterator(self):
+        v = reduce(lambda acc, el: acc.assoc(el, el+1), range(1500), Vector())
+        self.assertEqual(list(range(1, 1501)), list(v))
+        self.assertEqual(1125750, sum(v))
+
+    def test_index_error(self):
+        v = reduce(lambda acc, el: acc.assoc(el, el+2), range(50), Vector())
+        self.assertRaises(IndexError, v.get, -1)
+        self.assertRaises(IndexError, v.get, 50)
+        self.assertRaises(IndexError, v.get, 52)
+
+    def test_setitem_should_not_be_implemented(self):
+        def f():
+            v = Vector().cons(20)
+            v[0] = 10
+        self.assertRaises(NotImplementedError, f)
+
+    def test_subvector_operation(self):
+        pass
+
+class FingerTreeDequeTestCase(unittest.TestCase):
+    
+    def test_deque_basic_operations(self):
+        d1 = Deque()
+        d2 = d1.push_back(1)
+        d3 = d2.push_back(2)
+        d4 = d3.push_back(3)
+        d5 = d4.push_front(10)
+        d6 = d5.push_front(20)
+        self.assertEqual(1, d4.head())
+        self.assertEqual(3, d4.last())
+        self.assertEqual(20, d6.head())
+        self.assertEqual(3, d6.last())
+
+    def test_deque_num_of_elements(self):
+        pass
+
+    def test_deque_is_empty(self):
+        self.assertTrue(Deque().is_empty())
+        self.assertFalse(Deque().push_back(1).is_empty())
+        self.assertTrue(Deque().push_back(1).tail().is_empty())
+
+    def test_iterator(self):
+        self.assertEqual([], list(Deque()))
+        self.assertEqual([1,2,3], list(Deque().push_back(1).push_back(2).push_back(3)))
+        self.assertEqual(60, sum(Deque().push_back(10).push_front(20).push_back(30)))
+        self.assertEqual(sum(range(1,20)), sum(Deque.from_iterable(range(1,20))))
+
 if __name__ == '__main__':
     unittest.main()
